from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import attr

AtomicType = str
Type = Union[AtomicType, 'ListType', 'SetType', 'MapType', 'ReferenceType']


@dataclass()
class ListType:
    valueType: Type


@dataclass()
class SetType:
    valueType: Type


@dataclass()
class MapType:
    keyType: Type
    valueType: Type


@dataclass()
class ReferenceType:
    module: str
    name: str


@dataclass()
class Namespace:
    name: str
    language: str


@dataclass()
class EnumMember:
    name: str
    value: int
    doc: str = ''


@dataclass()
class Enum:
    name: str
    members: List[EnumMember]
    doc: str = ''


@dataclass()
class Typedef:
    name: str
    type_: Type
    doc: str = ''


@dataclass()
class Field:
    key: int
    name: str
    type_: Type
    required: str = 'required'
    doc: str = ''
    default: Optional[Any] = None
    type: Optional[Any] = None


@dataclass()
class Struct:
    name: str
    isException: bool
    isUnion: bool
    fields: List[Field]
    doc: str = ''


@dataclass()
class Constant:
    name: str
    type_: Type
    value: Any
    doc: str = ''


@dataclass()
class Function:
    name: str
    oneway: bool
    returnType: Type
    arguments: List[Field]
    exceptions: List[Field]
    doc: str = ''


@dataclass()
class Service:
    name: str
    functions: List[Function]
    doc: str = ''


@dataclass()
class Module:
    name: str
    namespaces: List[Namespace]
    enums: List[Enum]
    typedefs: List[Typedef]
    structs: List[Struct]
    constants: List[Constant]
    services: List[Service]
    doc: str = ''



import enum
import re
from dataclasses import dataclass
from itertools import groupby
from typing import (Any, Callable, Dict, Iterable, List, Optional, Tuple,
                    Union, cast)

from docutils import nodes
from docutils.parsers.rst.directives import flag, unchanged, unchanged_required
from sphinx import addnodes
from sphinx.addnodes import (desc_addname, desc_annotation, desc_content,
                             desc_name, desc_signature, desc_type,
                             pending_xref)
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, Index, IndexEntry, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.ext.autodoc import Documenter
from sphinx.locale import _
from sphinx.roles import XRefRole
from sphinx.util import docfields, logging
from sphinx.util.nodes import make_refnode

list_re = re.compile(r'^(list|set)<(.*)>$')
map_re = re.compile(r'^map<(.*),(.*)>$')


def make_desc_type(content: str) -> desc_type:
    return desc_type(content, content)


def parse_type(typename: str,
               inner_node: Callable[[str], nodes.Node]) -> List[nodes.Node]:
    typename = typename.replace(' ', '')
    m = list_re.match(typename)
    if m:
        return [inner_node(m.group(1) + '<')] + parse_type(
            m.group(2), inner_node) + [inner_node('>')]
    m = map_re.match(typename)
    if m:
        return [inner_node('map<')] + parse_type(m.group(1), inner_node) + [
            inner_node(', ')
        ] + parse_type(m.group(2), inner_node) + [inner_node('>')]
    return [
        pending_xref('',
                     inner_node(typename),
                     refdomain='thrift',
                     refexplicit=False,
                     reftarget=typename,
                     reftype='field')
    ]


class ThriftObjectKind(enum.Enum):
    Module = 'module'
    Constant = 'constant'
    TypeDef = 'typedef'
    Enum = 'enum'
    EnumField = 'enum_field'
    Struct = 'struct'
    StructField = 'struct_field'
    Exception = 'exception'
    ExceptionField = 'exception_field'
    Service = 'service'
    ServiceMethod = 'service_method'


@dataclass(frozen=True, unsafe_hash=True)
class Signature:
    kind: ThriftObjectKind
    name: str
    module: Optional[str]

    def __str__(self) -> str:
        if self.module is None:
            return f'{self.name}:{self.kind.value}'
        return f'{self.module}.{self.name}:{self.kind.value}'


@dataclass(frozen=True)
class ThriftObjectData:
    signature: Signature
    document: str


def _index_entry(
    object_description: str
) -> Callable[['ThriftDomain', str, str], Tuple[str, str, str, str,
                                                Optional[str]]]:
    def impl(domain: 'ThriftDomain', title: str,
             target: str) -> Tuple[str, str, str, str, Optional[str]]:
        return 'pair', f'{domain.object_types[object_description].lname} ; {title}', target, '', None

    return impl


_index_objects = {kind: _index_entry(kind.value) for kind in ThriftObjectKind}


class ThriftObject(ObjectDescription[Signature]):
    @property
    def kind(self) -> ThriftObjectKind:
        return ThriftObjectKind(self.objtype)

    def add_target_and_index(self, name: Signature, sig: str,
                             signode: desc_signature) -> None:
        if name not in self.state.document.ids:
            signode['names'].append(name)
            signode['ids'].append(name)
            signode['first'] = (not self.names)

            domain = cast(ThriftDomain, self.env.get_domain(self.domain))
            make_index = _index_objects.get(name.kind, None)
            if make_index is not None:
                self.indexnode['entries'].append(
                    make_index(domain, name.name, str(name)))
            self.state.document.note_explicit_target(signode)
            domain.add_thrift_object(ThriftObjectData(name, self.env.docname))


class ThriftModule(ThriftObject):
    def handle_signature(self, sig: str, signode: Any) -> Signature:
        signode += desc_annotation('module ', 'module ')
        signode += desc_name(sig, sig)
        return Signature(self.kind, sig, None)


class ThriftConstant(ThriftObject):
    required_arguments = 1
    option_spec = {'module': unchanged_required, 'type': unchanged_required}

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        signode += desc_annotation(self.objtype, self.objtype)
        module_name = self.options['module'] + '.'
        signode += desc_name(sig, sig)
        signode += desc_type(': ', ': ')
        signode.extend(parse_type(self.options['type'], make_desc_type))
        return Signature(self.kind, sig, self.options['module'])


class ThriftTypedef(ThriftObject):
    required_arguments = 1
    option_spec = {'module': unchanged_required, 'target': unchanged_required}

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        signode += desc_annotation(self.objtype, self.objtype)
        module_name = self.options['module'] + '.'
        signode += desc_name(sig, sig)
        signode += desc_type(' = ', ' = ')
        signode.extend(parse_type(self.options['target'], make_desc_type))
        return Signature(self.kind, sig, self.options['module'])


class ThriftEnum(ThriftObject):
    required_arguments = 1
    option_spec = {'module': unchanged_required}

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        signode += desc_annotation(self.objtype, self.objtype)
        module_name = self.options['module'] + '.'
        signode += desc_name(sig, sig)
        return Signature(self.kind, sig, self.options['module'])


class ThriftEnumField(ThriftObject):
    option_spec = {
        'module': unchanged_required,
        'enum': unchanged_required,
        'value': unchanged_required
    }

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        enum_name = self.options['enum'] + '.'
        signode += desc_addname(enum_name, enum_name)
        signode += desc_name(sig, sig)
        signode += desc_annotation(' = ' + self.options['value'],
                                   ' = ' + self.options['value'])
        return Signature(self.kind, enum_name + sig, self.options['module'])


class ThriftStruct(ThriftObject):
    required_arguments = 1
    option_spec = {'module': unchanged_required, 'exception': flag}

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        annotation = 'exception' if 'exception' in self.options else self.objtype
        signode += desc_annotation(annotation, annotation)
        module_name = self.options['module'] + '.'
        signode += desc_name(sig, sig)
        return Signature(self.kind, sig, self.options['module'])


class ThriftStructField(ThriftObject):
    required_arguments = 1
    option_spec = {
        'module': unchanged_required,
        'struct': unchanged_required,
        'type': unchanged_required
    }

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        annotation = 'field'
        signode += desc_annotation(annotation, annotation)
        struct_name = self.options['struct'] + '.'
        signode += desc_name(sig, sig)
        signode += desc_type(': ', ': ')
        signode.extend(parse_type(self.options['type'], make_desc_type))
        return Signature(self.kind, struct_name + sig, self.options['module'])


class ThriftService(ThriftObject):
    required_arguments = 1
    option_spec = {'module': unchanged_required}

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        annotation = 'service'
        signode += desc_annotation(annotation, annotation)
        module_name = self.options['module'] + '.'
        # signode += desc_addname(module_name, module_name)
        signode += desc_name(sig, sig)
        return Signature(self.kind, sig, self.options['module'])


def parameter_list(arg: str) -> List[Tuple[str, str]]:
    if arg is None:
        return []
    return [(p.split(';')[0], p.split(';')[1]) for p in arg.split()]


class ThriftServiceMethod(ThriftObject):
    required_arguments = 1
    option_spec = {
        'module': unchanged_required,
        'service': unchanged_required,
        'parameters': parameter_list,
        'exceptions': parameter_list,
        'return_type': unchanged_required,
        'oneway': flag
    }

    doc_field_types = [
        docfields.TypedField('parameter',
                             label='Parameters',
                             names=('param', ),
                             typenames=('type', ),
                             typerolename='field',
                             can_collapse=True),
        docfields.TypedField('exception',
                             label='Exceptions',
                             names=('throws', ),
                             typenames=('type', ),
                             typerolename='field',
                             can_collapse=True)
    ]

    def handle_signature(self, sig: str, signode: desc_signature) -> Signature:
        if 'oneway' in self.options:
            signode += desc_annotation('oneway', 'oneway')
        signode += desc_type(self.options['return_type'] + ' ',
                             self.options['return_type'] + ' ')
        service_name = self.options['service'] + '.'
        signode += desc_name(sig, sig)
        signode += desc_addname('(', '(')
        first = True
        for name, type_ in self.options['parameters']:
            if first:
                first = False
            else:
                signode += desc_addname(', ', ', ')
            signode.extend(parse_type(type_, make_desc_type))
            signode += make_desc_type(' ')
            signode += desc_addname(name, name)
        signode += desc_addname(')', ')')
        first = True
        if self.options['exceptions']:
            signode += desc_addname(' throws (', ' throws (')
        for name, type_ in self.options['exceptions']:
            if first:
                first = False
            else:
                signode += desc_addname(', ', ', ')
            signode.extend(parse_type(type_, make_desc_type))
            signode += make_desc_type(' ')
            signode += desc_addname(name, name)
        if self.options['exceptions']:
            signode += desc_addname(')', ')')
        return Signature(self.kind, service_name + sig, self.options['module'])


class ThriftXRefRole(XRefRole):
    def result_nodes(
            self, document: nodes.document, env: BuildEnvironment,
            node: nodes.Element, is_ref: bool
    ) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        if not is_ref:
            return [node], []

        make_index = _index_objects.get(ThriftObjectKind(node['reftype']),
                                        None)
        if make_index is not None:
            target_id = 'index-{}-{}'.format(
                env.new_serialno('index-{}'.format(node['reftarget'])),
                node['reftarget'])
            target = nodes.target('', '', ids=[target_id])
            document.note_explicit_target(target)

            index = addnodes.index()
            index['entries'] = [
                make_index(cast(ThriftDomain, env.get_domain('thrift')),
                           node['reftarget'], target_id)
            ]
            return [target, index, node], []
        return [node], []


class ThriftIndex(Index):
    name = 'modindex'
    localname = 'Thrift Index'
    shortname = 'Index'

    def generate(
        self,
        docnames: Iterable[str] = None
    ) -> Tuple[List[Tuple[str, List[IndexEntry]]], bool]:
        entries = []
        objects = cast(ThriftDomain, self.domain).thrift_objects
        for obj in objects:
            entries.append(
                IndexEntry(name=obj.signature.name,
                           subtype=0,
                           docname=obj.document,
                           anchor=str(obj.signature),
                           extra=self.domain.object_types[
                               obj.signature.kind.value].lname,
                           qualifier='',
                           descr=''))
        content: Dict[str, List[IndexEntry]] = {}
        for key, group in groupby(entries, key=lambda t: t[0][0].upper()):
            content.setdefault(key, []).extend(group)
        return sorted(content.items(), key=lambda t: t[0]), False


class ThriftDomain(Domain):
    name = 'thrift'
    object_types = {
        'module':
        ObjType(_('module'), 'module'),
        'namespace':
        ObjType(_('namespace'), 'namespace'),
        'constant':
        ObjType(_('constant'), 'constant'),
        'typedef':
        ObjType(_('type alias'), 'typedef'),
        'enum':
        ObjType(_('enumeration'), 'enum'),
        'enum_field':
        ObjType(_('enumerator'), 'enum_field'),
        'struct':
        ObjType(_('struct'), 'struct'),
        'struct_field':
        ObjType(_('struct field'), 'struct_field'),
        'exception':
        ObjType(_('exception'), 'struct', 'exception'),
        'exception_field':
        ObjType(_('exception field'), 'struct_field', 'exception_field'),
        'service':
        ObjType(_('service'), 'service'),
        'service_method':
        ObjType(_('service method'), 'service_method')
    }
    directives = {
        'module': ThriftModule,
        'constant': ThriftConstant,
        'typedef': ThriftTypedef,
        'enum': ThriftEnum,
        'enum_field': ThriftEnumField,
        'struct': ThriftStruct,
        'struct_field': ThriftStructField,
        'exception': ThriftStruct,
        'exception_field': ThriftStructField,
        'service': ThriftService,
        'service_method': ThriftServiceMethod
    }
    roles = {
        'module': ThriftXRefRole(),
        'constant': ThriftXRefRole(),
        'typedef': ThriftXRefRole(),
        'enum': ThriftXRefRole(),
        'enum_field': ThriftXRefRole(),
        'struct': ThriftXRefRole(),
        'struct_field': ThriftXRefRole(),
        'exception': ThriftXRefRole(),
        'exception_field': ThriftXRefRole(),
        'service': ThriftXRefRole(),
        'service_method': ThriftXRefRole()
    }
    indices = [ThriftIndex]
    initial_data: Dict[str, Any] = {'objects': []}

    def resolve_xref(self, env: BuildEnvironment, fromdocname: str,
                     builder: Builder, typ: str, target: str,
                     node: pending_xref,
                     contnode: nodes.Element) -> nodes.Element:
        def find_target() -> Optional[ThriftObjectData]:
            moduleless_results = []
            for obj in self.thrift_objects:
                if target.split('.') == [
                        obj.signature.module, obj.signature.name
                ]:
                    return obj
                if obj.signature.name == target:
                    moduleless_results.append(obj)
            if len(moduleless_results) == 1:
                return moduleless_results[0]
            elif len(moduleless_results) > 1:
                logger = logging.getLogger(__name__)
                logger.warning(
                    f'{target} is ambiguous. Possible resolutions:',
                    extra={
                        'resolutions':
                        list(str(obj.signature) for obj in moduleless_results)
                    })
            return None

        targetdata = find_target()
        if targetdata is not None:
            return make_refnode(builder, fromdocname, targetdata.document,
                                str(targetdata.signature), contnode)
        else:
            return None

    def merge_domaindata(self, docnames: List[str], otherdata: Dict) -> None:
        objects = cast(Iterable[ThriftObject], otherdata['objects'])
        for obj in objects:
            if obj.docname in docnames:
                self.add_thrift_object(obj)

    def add_thrift_object(self, obj: ThriftObjectData) -> None:
        cast(List[ThriftObjectData], self.data['objects']).append(obj)

    @property
    def thrift_objects(self) -> Iterable[ThriftObjectData]:
        return self.data['objects']



from typing import Any, Dict

from sphinx.application import Sphinx

__version__ = '0.1.0'


def setup(app: Sphinx) -> Dict[str, Any]:
    from sphinx_thrift.documenter import ThriftModuleDocumenter
    from sphinx_thrift.domain import ThriftDomain

    app.add_config_value('thrift_executable', 'thrift', 'env', [str])
    app.add_autodocumenter(ThriftModuleDocumenter)
    app.add_domain(ThriftDomain)
    return {
        'version': __version__,
        'env_version': 0,
        'parallel_read_safe': True,
        'parallel_write_safe': True
    }



from subprocess import check_call
from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

from sphinx.errors import ConfigError
from sphinx.ext.autodoc import Documenter, ModuleDocumenter
from sphinx.ext.autodoc.directive import DocumenterBridge

import sphinx_thrift.thrift_ast as ast
from sphinx_thrift.thrift_ast import (Constant, Enum, Function, Service,
                                      Struct, Typedef)


def typeId(type_: ast.Type) -> str:
    dispatch = {
        ast.ListType:
        lambda p: 'list<{}>'.format(typeId(p.valueType)),
        ast.SetType:
        lambda p: 'set<{}>'.format(typeId(p.valueType)),
        ast.MapType:
        lambda p: 'map<{},{}>'.format(typeId(p.keyType), typeId(p.valueType)),
        ast.ReferenceType:
        lambda r: (r.module + '.' if r.module else '') + r.name,
        str:
        lambda s: s.strip()
    }
    return dispatch[type_.__class__](type_)


class _DirectiveGenerator:
    def __init__(self,
                 add_line: Callable[[str], None],
                 directive: str,
                 indent: int = 0) -> None:
        self._add_line = add_line
        self.directive = directive
        self.indent = indent * ' '

    def add_line(self, content: str) -> None:
        self._add_line(self.indent + content)

    def add_header(self, name: str, attributes: Dict[str, str]) -> None:
        self.add_line('.. thrift:{}:: {}'.format(self.directive, name))
        for item in attributes.items():
            self.add_line('   :{}: {}'.format(*item))
        self.add_line('')

    def add_fields(
            self,
            fields: Iterable[Tuple[str, Union[str, Tuple[str, str]]]]) -> None:
        for name, value in fields:
            if isinstance(value, str):
                self.add_line('   :{}: {}'.format(name, value))
            else:
                self.add_line('   :{} {}: {}'.format(name, *value))
        self.add_line('')

    def add_doc(self, doc: str) -> None:
        self.add_line('   ' + doc)
        self.add_line('')

    def generate(
        self,
        name: str,
        doc: str,
        attributes: Dict[str, str] = {},
        fields: Iterable[Tuple[str, Union[str, Tuple[str,
                                                     str]]]] = []) -> None:
        self.add_header(name, attributes)
        self.add_fields(fields)
        self.add_doc(doc)


class ThriftDocumenter(Documenter):
    objtype = 'object'
    titles_allowed = True


class ThriftModuleDocumenter(ThriftDocumenter):
    objtype = 'thrift_module'
    module: ast.Module  # type: ignore

    def __init__(self,
                 directive: DocumenterBridge,
                 name: str,
                 indent: str = '') -> None:
        super().__init__(directive, name, indent)
        self.thrift_executable = self.env.config['thrift_executable']
        self.filename = self.name + '.thrift'
        self.module_generator = self._create_generator('module')
        self.service_generator = self._create_generator('service')
        self.method_generator = self._create_generator('service_method',
                                                       indent=3)
        self.enum_generator = self._create_generator('enum')
        self.enum_field_generator = self._create_generator('enum_field',
                                                           indent=3)
        self.struct_generator = self._create_generator('struct')
        self.struct_field_generator = self._create_generator('struct_field',
                                                             indent=3)
        self.constant_generator = self._create_generator('constant')
        self.typedef_generator = self._create_generator('typedef')

    def _create_generator(self,
                          name: str,
                          indent: int = 0) -> _DirectiveGenerator:
        return _DirectiveGenerator(self._add_line, name, indent)

    @classmethod
    def can_document_member(cls, member: Any, membername: str, isattr: bool,
                            parent: Any) -> bool:
        """Called to see if a member can be documented by this documenter."""
        return False

    def get_sourcename(self) -> str:
        return f'{self.filename}:docstring of {self.name}'

    def _add_line(self, content: str) -> None:
        self.add_line(content, self.get_sourcename())

    def generate(self,
                 more_content: Any = None,
                 real_modname: str = None,
                 check_module: bool = False,
                 all_members: bool = False) -> None:
        import os.path
        import shutil

        from sphinx_thrift.parser import load_module

        base_name = os.path.basename(self.name)
        abs_thrift_path = self.filename if os.path.isabs(
            self.filename) else os.path.join(
                os.path.dirname(self.env.doc2path(self.env.docname)),
                self.filename)
        self.env.note_dependency(abs_thrift_path)

        if shutil.which(self.thrift_executable) is None:
            raise ConfigError("Couldn't find thrift executable")
        check_call([
            self.thrift_executable, '--gen', 'xml', '--out',
            self.env.doctreedir, abs_thrift_path
        ])
        self.module = load_module(f'{self.env.doctreedir}/{base_name}.xml')
        self.module_generator.generate(self.module.name,
                                       self.module.doc,
                                       fields=[(ns.language,
                                                ':code:`' + ns.name + '`')
                                               for ns in self.module.namespaces
                                               ])
        self._generate_constants()
        self._generate_typedefs()
        self._generate_enums()
        self._generate_structs()
        self._generate_services()

    def _generate_constants(self) -> None:
        if not self.module.constants:
            return
        self._add_line('Constants')
        self._add_line('---------')
        for cons in self.module.constants:
            self.constant_generator.generate(cons.name, cons.doc, {
                'module': self.module.name,
                'type': typeId(cons.type_)
            })

    def _generate_typedefs(self) -> None:
        if not self.module.typedefs:
            return
        self._add_line('Type aliases')
        self._add_line('------------')
        for td in self.module.typedefs:
            self.typedef_generator.generate(td.name, td.doc, {
                'module': self.module.name,
                'target': typeId(td.type_)
            })

    def _generate_enum(self, enum: Enum) -> None:
        self.enum_generator.generate(enum.name,
                                     enum.doc,
                                     attributes={'module': self.module.name})
        for member in enum.members:
            self.enum_field_generator.generate(member.name,
                                               member.doc,
                                               attributes={
                                                   'module': self.module.name,
                                                   'enum': enum.name,
                                                   'value': str(member.value)
                                               })

    def _generate_enums(self) -> None:
        if not self.module.enums:
            return
        self._add_line('Enumerations')
        self._add_line('------------')
        for enum in self.module.enums:
            self._generate_enum(enum)

    def _generate_struct(self, struct: Struct) -> None:
        attributes = {'module': self.module.name}
        if struct.isException:
            attributes['exception'] = ''
        self.struct_generator.generate(struct.name,
                                       struct.doc,
                                       attributes=attributes)
        for m in struct.fields:
            member_attrs = {
                'module': self.module.name,
                'struct': struct.name,
                'type': typeId(m.type_)
            }
            fields = {}
            if m.default is not None:
                fields['default'] = m.default
            self.struct_field_generator.generate(m.name,
                                                 m.doc,
                                                 attributes=member_attrs,
                                                 fields=fields.items())

    def _generate_structs(self) -> None:
        if not self.module.structs:
            return
        self._add_line('Structs')
        self._add_line('-------')
        for struct in self.module.structs:
            self._generate_struct(struct)

    def _generate_method(self, service: Service, method: Function) -> None:
        params = ' '.join(f'{p.name};{typeId(p.type_)}'
                          for p in method.arguments)
        exceptions = ' '.join(f'{p.name};{typeId(p.type_)}'
                              for p in method.exceptions)
        attributes = {
            'module': self.module.name,
            'service': service.name,
            'return_type': typeId(method.returnType),
            'parameters': params,
            'exceptions': exceptions
        }
        if method.oneway:
            attributes['oneway'] = ''
        fields = []
        for p in method.arguments:
            fields.append(('param', (p.name, p.doc)))
        for e in method.exceptions:
            fields.append(('throws', (e.name, e.doc)))
        self.method_generator.generate(method.name,
                                       method.doc,
                                       attributes=attributes,
                                       fields=fields)

    def _generate_service(self, service: Service) -> None:
        self.service_generator.generate(
            service.name, service.doc, attributes={'module': self.module.name})
        for method in service.functions:
            self._generate_method(service, method)

    def _generate_services(self) -> None:
        if not self.module.services:
            return
        self._add_line('Services')
        self._add_line('--------')
        for service in self.module.services:
            self._generate_service(service)



import xml.etree.ElementTree as ET
from enum import Enum
from typing import Callable, Dict, List

import sphinx_thrift.thrift_ast as ast


class Tag(Enum):
    NAMESPACE = 'namespace'
    CONSTANT = 'const'
    TYPEDEF = 'typedef'
    ENUM = 'enum'
    FIELD = 'field'
    ARG = 'arg'
    STRUCT = 'struct'
    EXCEPTION = 'exception'
    SERVICE = 'service'
    METHOD = 'method'
    THROWS = 'throws'


def parse_list_type(root: ET.Element) -> ast.ListType:
    value_node = root.find('elemType')
    assert (value_node is not None)
    return ast.ListType(valueType=parse_type(value_node))


def parse_set_type(root: ET.Element) -> ast.SetType:
    value_node = root.find('elemType')
    assert (value_node is not None)
    return ast.SetType(valueType=parse_type(value_node))


def parse_map_type(root: ET.Element) -> ast.MapType:
    key_node = root.find('keyType')
    value_node = root.find('valueType')
    assert (key_node is not None)
    assert (value_node is not None)
    return ast.MapType(keyType=parse_type(key_node),
                       valueType=parse_type(value_node))


def parse_reference_type(root: ET.Element) -> ast.ReferenceType:
    return ast.ReferenceType(module=root.attrib['type-module'],
                             name=root.attrib['type-id'])


def parse_type(root: ET.Element) -> ast.Type:
    type_funcs: Dict[str, Callable[[ET.Element], ast.Type]] = {
        'list': parse_list_type,
        'set': parse_set_type,
        'map': parse_map_type,
        'id': parse_reference_type
    }
    t = root.attrib['type']
    return type_funcs.get(t, lambda _: t)(root)


def parse_namespace(root: ET.Element) -> ast.Namespace:
    assert (root.tag == Tag.NAMESPACE.value)
    return ast.Namespace(name=root.attrib['value'],
                         language=root.attrib['name'])


def parse_constant(root: ET.Element) -> ast.Constant:
    assert (root.tag == Tag.CONSTANT.value)
    return ast.Constant(name=root.attrib['name'],
                        doc=root.attrib.get('doc', ''),
                        type_=parse_type(root),
                        value=None)


def parse_typedef(root: ET.Element) -> ast.Typedef:
    assert (root.tag == Tag.TYPEDEF.value)
    return ast.Typedef(name=root.attrib['name'],
                       doc=root.get('doc', ''),
                       type_=parse_type(root))


def parse_enum(root: ET.Element) -> ast.Enum:
    def parse_members() -> List[ast.EnumMember]:
        return [
            ast.EnumMember(name=n.attrib['name'],
                           doc=n.attrib.get('doc', ''),
                           value=int(n.attrib['value'])) for n in root
        ]

    assert (root.tag == Tag.ENUM.value)
    return ast.Enum(name=root.attrib['name'],
                    doc=root.attrib.get('doc', ''),
                    members=parse_members())


def _parse_field_with_tag(tag: str, root: ET.Element) -> ast.Field:
    assert (root.tag == tag)
    return ast.Field(name=root.attrib['name'],
                     key=int(root.attrib['field-id']),
                     doc=root.get('doc', ''),
                     required=root.get('required', 'required'),
                     type_=parse_type(root))


def parse_field(root: ET.Element) -> ast.Field:
    return _parse_field_with_tag(Tag.FIELD.value, root)


def parse_arg(root: ET.Element) -> ast.Field:
    return _parse_field_with_tag(Tag.ARG.value, root)


def parse_throws(root: ET.Element) -> ast.Field:
    return _parse_field_with_tag(Tag.THROWS.value, root)


def parse_struct(root: ET.Element) -> ast.Struct:
    assert (root.tag == Tag.STRUCT.value)
    return ast.Struct(
        name=root.attrib['name'],
        doc=root.get('doc', ''),
        isException=False,
        isUnion=False,
        fields=[parse_field(f) for f in root if f.tag == Tag.FIELD.value])


def parse_exception(root: ET.Element) -> ast.Struct:
    assert (root.tag == Tag.EXCEPTION.value)
    return ast.Struct(
        name=root.attrib['name'],
        doc=root.get('doc', ''),
        isException=True,
        isUnion=False,
        fields=[parse_field(f) for f in root if f.tag == Tag.FIELD.value])


def parse_method(root: ET.Element) -> ast.Function:
    assert (root.tag == Tag.METHOD.value)
    returns = root.find('returns')
    assert (returns is not None)
    return ast.Function(
        name=root.attrib['name'],
        doc=root.get('doc', ''),
        oneway=root.get('oneway', 'false') == 'true',
        returnType=parse_type(returns),
        arguments=list(map(parse_arg, root.findall(Tag.ARG.value))),
        exceptions=list(map(parse_throws, root.findall(Tag.THROWS.value))))


def parse_service(root: ET.Element) -> ast.Service:
    assert (root.tag == Tag.SERVICE.value)
    return ast.Service(name=root.attrib['name'],
                       doc=root.attrib.get('doc', ''),
                       functions=[parse_method(m) for m in root])


def parse_module(root: ET.Element) -> ast.Module:
    assert (root.tag == 'document')
    return ast.Module(
        name=root.attrib['name'],
        doc=root.attrib.get('doc', ''),
        namespaces=[
            parse_namespace(ns) for ns in root.findall(Tag.NAMESPACE.value)
        ],
        typedefs=[parse_typedef(td) for td in root.findall(Tag.TYPEDEF.value)],
        constants=[
            parse_constant(const) for const in root.findall(Tag.CONSTANT.value)
        ],
        enums=[parse_enum(enum) for enum in root.findall(Tag.ENUM.value)],
        structs=[
            parse_struct(struct) for struct in root.findall(Tag.STRUCT.value)
        ] +
        [parse_exception(exc) for exc in root.findall(Tag.EXCEPTION.value)],
        services=[
            parse_service(service)
            for service in root.findall(Tag.SERVICE.value)
        ])


def load_module(filename: str) -> ast.Module:
    et = ET.iterparse(filename)
    for _, el in et:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]  # strip all namespaces
        for at in el.attrib.keys():  # strip namespaces of attributes too
            if '}' in at:
                newat = at.split('}', 1)[1]
                el.attrib[newat] = el.attrib[at]
                del el.attrib[at]
    root = et.root  # type: ignore
    docu = root.find('document')
    assert (docu is not None)
    return parse_module(docu)
